import zope.interface, logging, time, bcrypt, json, os
from datetime import datetime

from Houdini.Plugins import Plugin
from Houdini.Handlers import Handlers
from Houdini.Data.Penguin import Penguin, BuddyList
from Houdini.Data.Ban import Ban
from Houdini.Crypto import Crypto
from Houdini.Handlers.Login.Login import handleLogin

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet import threads

from sqlalchemy.sql import select

import treq

class Captcha(object):
    zope.interface.implements(Plugin)

    author = "Ben"
    version = 0.1
    description = "Google recaptcha support in Houdini"

    def __init__(self, server):
        self.logger = logging.getLogger("Houdini")

        self.server = server

        configFile = os.path.dirname(os.path.realpath(__file__)) + "/captcha.conf"
        with open(configFile, "r") as fileHandle:
            self.config = json.load(fileHandle)

        self.verifyRecaptchaUrl = self.config["VerifyUrl"]
        self.secretRecaptchaKey = self.config["SecretKey"]
        self.recaptchaHost = self.config["Hostname"]

        Handlers.Login -= handleLogin
        Handlers.Login += self.handleLogin

    @inlineCallbacks
    def handleLogin(self, player, data):
        if player.randomKey is None:
            returnValue(player.transport.loseConnection())

        if not hasattr(self.server, "loginAttempts"):
            self.server.loginAttempts = {}

        loginTimestamp = time.time()
        username = data.Username
        password = data.Password[:32]
        captchaToken = data.Password[32:]

        self.logger.info("{0} is attempting to login..".format(username))

        user = yield player.engine.first(Penguin.select(Penguin.c.Username == username))

        if user is None:
            returnValue(player.sendErrorAndDisconnect(100))

        ipAddr = player.transport.getPeer().host

        passwordCorrect = yield threads.deferToThread(bcrypt.checkpw, password, user.Password)

        if not passwordCorrect:
            self.logger.info("{} failed to login.".format(username))

            if ipAddr in self.server.loginAttempts:
                lastFailedAttempt, failureCount = self.server.loginAttempts[ipAddr]

                failureCount = 1 if loginTimestamp - lastFailedAttempt >= self.server.server["LoginFailureTimer"] \
                    else failureCount + 1

                self.server.loginAttempts[ipAddr] = [loginTimestamp, failureCount]

                if failureCount >= self.server.server["LoginFailureLimit"]:
                    returnValue(player.sendErrorAndDisconnect(150))

            else:
                self.server.loginAttempts[ipAddr] = [loginTimestamp, 1]

            returnValue(player.sendErrorAndDisconnect(101))

        if ipAddr in self.server.loginAttempts:
            previousAttempt, failureCount = self.server.loginAttempts[ipAddr]

            maxAttemptsExceeded = failureCount >= self.server.server["LoginFailureLimit"]
            timerSurpassed = (loginTimestamp - previousAttempt) > self.server.server["LoginFailureTimer"]

            if maxAttemptsExceeded and not timerSurpassed:
                returnValue(player.sendErrorAndDisconnect(150))
            else:
                del self.server.loginAttempts[ipAddr]

        postData = {
            "secret": self.secretRecaptchaKey,
            "response": captchaToken,
            "remoteip": ipAddr
        }

        googleResponse = yield treq.get(self.verifyRecaptchaUrl, params=postData)
        captchaResult = yield googleResponse.json()

        if not captchaResult["success"]:
            returnValue(player.sendErrorAndDisconnect(101))

        if bool(self.recaptchaHost) and captchaResult["hostname"] != self.recaptchaHost:
            returnValue(player.sendErrorAndDisconnect(101))

        if not user.Active:
            returnValue(player.sendErrorAndDisconnect(900))

        if user.Permaban:
            returnValue(player.sendErrorAndDisconnect(603))

        activeBan = yield player.engine.first(
            Ban.select((Ban.c.PenguinID == user.ID) & (Ban.c.Expires >= datetime.now())))

        if activeBan is not None:
            hoursLeft = round((activeBan.Expires - datetime.now()).total_seconds() / 60 / 60)

            if hoursLeft == 0:
                returnValue(player.sendErrorAndDisconnect(602))

            else:
                player.sendXt("e", 601, hoursLeft)
                returnValue(player.transport.loseConnection())

        self.logger.info("{} logged in successfully".format(username))

        randomKey = Crypto.generateRandomKey()
        loginKey = Crypto.hash(randomKey[::-1])

        player.user = user
        player.engine.execute(Penguin.update(Penguin.c.ID == user.ID).values(LoginKey=loginKey))

        buddyWorlds = []
        worldPopulations = []

        serversConfig = self.server.config["Servers"]

        for serverName in serversConfig.keys():
            if serversConfig[serverName]["World"]:
                serverPopulation = yield threads.deferToThread(self.server.redis.get,
                                                               "{}.population".format(serverName))

                if not serverPopulation is None:
                    serverPopulation = int(serverPopulation) / (serversConfig[serverName]["Capacity"] / 6)
                else:
                    serverPopulation = 0

                serverPlayers = yield threads.deferToThread(self.server.redis.smembers, "{}.players".format(serverName))

                worldPopulations.append("%s,%s" % (serversConfig[serverName]["Id"], serverPopulation))

                if not len(serverPlayers) > 0:
                    self.logger.debug("Skipping buddy iteration for %s " % serverName)
                    continue

                buddies = yield player.engine.fetchall(
                    select([BuddyList.c.BuddyID]).where(BuddyList.c.PenguinID == player.user.ID))

                for buddyId, in buddies:
                    if str(buddyId) in serverPlayers:
                        buddyWorlds.append(serversConfig[serverName]["Id"])
                        break

        player.sendXt("l", user.ID, loginKey, "|".join(buddyWorlds), "|".join(worldPopulations))
        player.transport.loseConnection()

    def ready(self):
        self.logger.info("Google ReCAPTCHA plugin is ready!")